#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""Use command line to operate on datasets through click."""

import logging
import os
import sys
from configparser import ConfigParser
from pathlib import Path, PurePosixPath
from typing import Dict, Iterable, Iterator, List, Tuple, Union

import click

from .. import __version__
from ..dataset import Data, Segment
from ..utility import TBRN, TBRNType
from .dataset import DataSegmentClient, FusionDataSegmentClient
from .gas import GAS

_ACCESSKEY_PREFIX = "Accesskey-"


def _config_filepath() -> str:
    """Get the path of the config file.

    :return: the path of the config file
    """
    home = "HOMEPATH" if os.name == "nt" else "HOME"
    return os.path.join(os.environ[home], ".gasconfig")


def _read_config(config_filepath: str, config_name: str) -> Tuple[str, str]:
    """Read accessKey and url from the config file.

    :param config_filepath: the file containing config info
    :param config_name: the environment to login
    :returns:
        the accessKey of config_name read from the config file
        the url of config_name read from the config file
    """
    if not os.path.exists(config_filepath):
        click.echo(
            f"{config_filepath} not exist"
            "\n\nPlease use 'gas config <accessKey>' to create config file",
            err=True,
        )
        sys.exit(1)

    config_parser = ConfigParser()
    config_parser.read(config_filepath)
    access_key = config_parser[config_name]["accessKey"]
    url = config_parser[config_name]["url"] if "url" in config_parser[config_name] else ""
    return access_key, url


def _gas(access_key: str, url: str, config_name: str) -> GAS:
    """Load an object of class <GAS>.

    :param access_key: the accessKey of gas
    :param url: the login url
    :param config_name: the environment to login
        We will read accessKey and url from the appointed config_name and login gas.
    :return: the loaded gas client
    """
    if not access_key and not url:
        access_key, url = _read_config(_config_filepath(), config_name)

    if not access_key:
        click.echo("accessKey should be appointed", err=True)
        sys.exit(1)

    return GAS(access_key, url)


@click.group()
@click.version_option(version=__version__, message="%(version)s")
@click.option("-k", "--key", "access_key", type=str, default="", help="The accessKey of gas.")
@click.option("-u", "--url", type=str, default="", help="The login url.")
@click.option(
    "-c",
    "--config",
    "config_name",
    type=str,
    default="default",
    help="The environment to login.",
)
@click.option("-d", "--debug", is_flag=True, help="Debug mode.")
@click.pass_context
def cli(ctx: click.Context, access_key: str, url: str, config_name: str, debug: bool) -> None:
    """You can use 'gas' + COMMAND to operate on your dataset.\f

    :param ctx: the context to be passed as first argument
    :param access_key: the accessKey of gas
    :param url: the login url
    :param config_name: the environment to login
    :param debug: debug mode flag
    """
    ctx.obj = {
        "access_key": access_key,
        "url": url,
        "config_name": config_name,
    }

    if debug:
        logging.basicConfig(level=logging.DEBUG)


@cli.command()
@click.argument("name", type=str)
@click.pass_obj
def create(obj: Dict[str, str], name: str) -> None:
    """Create a dataset.\f

    :param obj: a dictionary including config info
    :param name: the name of the dataset to be created, like "tb:KITTI"
    """
    info = TBRN(tbrn=name)
    if info.type != TBRNType.DATASET:
        click.echo(f'"{name}" is not a dataset', err=True)
        sys.exit(1)
    _gas(**obj).create_dataset(info.dataset_name)


@cli.command()
@click.argument("name", type=str)
@click.pass_obj
def publish(obj: Dict[str, str], name: str) -> None:
    """Publish a dataset.\f

    :param obj: a dictionary including config info
    :param name: the name of the dataset to be published, like "tb:KITTI"
    """
    info = TBRN(tbrn=name)
    if info.type != TBRNType.DATASET:
        click.echo(f'"{name}" is not a dataset', err=True)
        sys.exit(1)
    dataset = _gas(**obj).get_dataset(info.dataset_name)
    dataset.publish()


@cli.command()
@click.argument("name", type=str)
@click.option("-y", "--yes", is_flag=True, help="Confirm to delete the dataset completely.")
@click.pass_obj
def delete(obj: Dict[str, str], name: str, yes: bool) -> None:
    """Delete a dataset.\f

    :param obj: a dictionary including config info
    :param name: the name of the dataset to be deleted, like "tb:KITTI"
    :param yes: confirm to delete the dataset completely
    """
    info = TBRN(tbrn=name)
    if info.type != TBRNType.DATASET:
        click.echo(f'"{name}" is not a dataset', err=True)
        sys.exit(1)

    if not yes:
        click.confirm(
            f'Dataset "{name}" will be completely deleted.\nDo you want to continue?', abort=True
        )

    _gas(**obj).delete_dataset(info.dataset_name)


@cli.command()
@click.argument("local_paths", type=str, nargs=-1)
@click.argument("tbrn", type=str, nargs=1)
@click.option(
    "-r", "--recursive", "is_recursive", is_flag=True, help="Copy directories recursively."
)
@click.option("-j", "--jobs", type=int, default=1, help="The number of threads.")
@click.option(
    "-s",
    "--skip_uploaded_files",
    "skip_uploaded_files",
    is_flag=True,
    help="Whether to skip the uploaded files.",
)
@click.pass_obj
def cp(  # pylint: disable=invalid-name, too-many-arguments
    obj: Dict[str, str],
    local_paths: Tuple[str],
    tbrn: str,
    is_recursive: bool,
    jobs: int,
    skip_uploaded_files: bool,
) -> None:
    """Copy local data to a remote path.\f

    :param obj: a dictionary including config info
    :param local_paths: a tuple of local paths containing data to be uploaded
    :param tbrn: the path to save the uploaded data, like "tb:KITTI:seg1"
    :param is_recursive: whether copy directories recursively
    :param jobs: number of threads to upload data
    :param skip_uploaded_files: Set it to `True` to skip the uploaded files
    """
    info = TBRN(tbrn=tbrn)

    if info.type == TBRNType.DATASET:
        click.echo("Error: SEGMENT is required", err=True)
        sys.exit(1)

    if info.type in [TBRNType.SEGMENT, TBRNType.NORMAL_FILE]:
        remote_path = info.remote_path if info.type == TBRNType.NORMAL_FILE else ""

        dataset = _gas(**obj).get_or_create_dataset(info.dataset_name)
        local_abspaths: List[str] = [os.path.abspath(local_path) for local_path in local_paths]
        if (
            len(local_abspaths) == 1
            and not os.path.isdir(local_abspaths[0])
            and remote_path
            and not remote_path.endswith("/")
        ):
            dataset.get_or_create_segment(info.segment_name).upload_data(
                local_abspaths[0], remote_path
            )
            return

        segment = _get_segment_object(info.segment_name, local_abspaths, remote_path, is_recursive)
        dataset.upload_segment_object(segment, jobs=jobs, skip_uploaded_files=skip_uploaded_files)
        return

    click.echo(f'"{tbrn}" is an invalid path', err=True)
    sys.exit(1)


def _get_segment_object(
    segment_name: str,
    local_abspaths: List[str],
    remote_path: str,
    is_recursive: bool,
) -> Segment:
    """Get the pair of local_path and remote_path.

    :param segment_name: the name of the segment these data belong to
    :param local_abspaths: a list of local abstract paths, could be folder or file
    :param remote_path: the remote object path, not necessarily end with '/'
    :param is_recursive: whether copy directories recursively
    :return: a segment containing mapping data
    """
    segment = Segment(segment_name)
    for local_abspath in local_abspaths:
        if not os.path.isdir(local_abspath):
            data = Data(
                local_abspath,
                remote_path=str(PurePosixPath(remote_path, os.path.basename(local_abspath))),
            )
            segment.append(data)
            continue

        if not is_recursive:
            click.echo(
                "Error: local paths include directories, please use -r option",
                err=True,
            )
            sys.exit(1)

        local_abspath = os.path.normpath(local_abspath)
        folder_name = os.path.basename(local_abspath)
        for root, _, filenames in os.walk(local_abspath):
            relpath = os.path.relpath(root, local_abspath) if root != local_abspath else ""
            for filename in filenames:
                data = Data(
                    os.path.join(root, filename),
                    remote_path=str(
                        PurePosixPath(Path(remote_path, folder_name, relpath, filename))
                    ),
                )
                segment.append(data)
    return segment


def _echo_segment(
    dataset_name: str,
    segment_name: str,
    segment: Union[DataSegmentClient, FusionDataSegmentClient],
    list_all_files: bool,
) -> None:
    """Echo a segment.

    :param dataset_name: the name of the dataset
    :param segment_name: the name of the segment
    :param segment: a normal/fusion segment
    :param list_all_files: only works when segment is a fusion one
        if False, list frame indexes only
        if True, list sensors and files, too
    """
    if isinstance(segment, DataSegmentClient):
        _echo_data(dataset_name, segment_name, segment.list_data())
    else:
        frames = segment.list_frame_objects()
        if not list_all_files:
            for index in range(len(frames)):
                click.echo(TBRN(dataset_name, segment_name, index).get_tbrn())
        else:
            for frame in frames:
                for data in frame.values():
                    click.echo(data.tbrn)


def _echo_data(dataset_name: str, segment_name: str, data_iter: Iterable[str]) -> None:
    """Echo files in data_iter under 'tb:dataset_name:segment_name'.

    :param dataset_name: the name of the dataset the segment belongs to
    :param segment_name: the name of the segment
    :param data_iter: iterable data to be echoed
    """
    for data in data_iter:
        click.echo(TBRN(dataset_name, segment_name, remote_path=data).get_tbrn())


def _ls_dataset(gas: GAS, info: TBRN, list_all_files: bool) -> None:
    dataset = gas._get_dataset(info.dataset_name)  # pylint: disable=protected-access
    segment_names = dataset.list_segments()
    if not list_all_files:
        for segment_name in segment_names:
            click.echo(TBRN(info.dataset_name, segment_name).get_tbrn())
        return

    for segment_name in segment_names:
        segment = dataset.get_segment(segment_name)
        _echo_segment(info.dataset_name, segment_name, segment, list_all_files)


def _ls_segment(gas: GAS, info: TBRN, list_all_files: bool) -> None:
    dataset = gas._get_dataset(info.dataset_name)  # pylint: disable=protected-access
    _echo_segment(
        info.dataset_name, info.segment_name, dataset.get_segment(info.segment_name), list_all_files
    )


def _ls_frame(gas: GAS, info: TBRN, list_all_files: bool) -> None:
    dataset = gas.get_fusion_dataset(info.dataset_name)
    fusion_segment = dataset.get_segment(info.segment_name)
    frame = fusion_segment.list_frame_objects()[info.frame_index]

    if not list_all_files:
        for sensor_name in frame:
            click.echo(TBRN(info.dataset_name, info.segment_name, info.frame_index, sensor_name))
    else:
        for data in frame.values():
            click.echo(data.tbrn)


def _ls_sensor(
    gas: GAS,
    info: TBRN,
    list_all_files: bool,  # pylint: disable=unused-argument
) -> None:
    dataset = gas.get_fusion_dataset(info.dataset_name)
    fusion_segment = dataset.get_segment(info.segment_name)
    frame = fusion_segment.list_frame_objects()[int(info.frame_index)]
    data = frame[info.sensor_name]
    click.echo(data.tbrn)


def _ls_fusion_file(
    gas: GAS,
    info: TBRN,
    list_all_files: bool,  # pylint: disable=unused-argument
) -> None:
    dataset = gas.get_fusion_dataset(info.dataset_name)
    fusion_segment = dataset.get_segment(info.segment_name)
    frame = fusion_segment.list_frame_objects()[info.frame_index]

    if frame[info.sensor_name].remote_path != info.remote_path:
        click.echo(f'No such file: "{info.remote_path}"!', err=True)
        sys.exit(1)

    click.echo(info)


def _ls_normal_file(  # pylint: disable=unused-argument
    gas: GAS, info: TBRN, list_all_files: bool
) -> None:
    dataset = gas.get_dataset(info.dataset_name)
    segment = dataset.get_segment(info.segment_name)
    _echo_data(
        info.dataset_name,
        info.segment_name,
        _filter_data(segment.list_data(), info.remote_path),
    )


_LS_FUNCS = {
    TBRNType.DATASET: _ls_dataset,
    TBRNType.SEGMENT: _ls_segment,
    TBRNType.NORMAL_FILE: _ls_normal_file,
    TBRNType.FRAME: _ls_frame,
    TBRNType.FRAME_SENSOR: _ls_sensor,
    TBRNType.FUSION_FILE: _ls_fusion_file,
}


@cli.command()
@click.argument("tbrn", type=str, default="")
@click.option(
    "-a", "--all", "list_all_files", is_flag=True, help="List all files under the segment."
)
@click.pass_obj
def ls(  # pylint: disable=invalid-name
    obj: Dict[str, str], tbrn: str, list_all_files: bool
) -> None:
    """List data under the path. If path is empty, list the names of all datasets.\f

    :param obj: a dictionary including config info
    :param tbrn: path to be listed, like "tb:KITTI:seg1". If empty, list names of all datasets.
    :param list_all_files: if list all files under the segment
    """
    gas = _gas(**obj)

    if not tbrn:
        for dataset_name in gas.list_datasets():
            click.echo(TBRN(dataset_name).get_tbrn())
        return

    info = TBRN(tbrn=tbrn)
    _LS_FUNCS[info.type](gas, info, list_all_files)


def _filter_data(
    data_list: List[str], remote_path: str, is_recursive: bool = True
) -> Iterator[str]:
    """Get a list of paths under the remote_path.

    :param data_list: a list of candidate paths
    :param remote_path: the remote path to filter data
    :param is_recursive: whether to filter data recursively
    :return: a list of paths under the given remote_path
    """
    if is_recursive:
        return (
            filter(lambda x: x.startswith(remote_path), data_list)
            if remote_path.endswith("/")
            else filter(lambda x: x.startswith(remote_path + "/") or x == remote_path, data_list)
        )
    return filter(lambda x: x == remote_path, data_list)


@cli.command()
@click.argument("tbrn", type=str)
@click.option(
    "-r", "--recursive", "is_recursive", is_flag=True, help="Remove directories recursively."
)
@click.option("-f", "--force", "force_delete", is_flag=True, help="Force to delete any segment.")
@click.pass_obj
# pylint: disable=invalid-name
def rm(obj: Dict[str, str], tbrn: str, is_recursive: bool, force_delete: bool) -> None:
    """Remove the remote paths.\f

    :param obj: a dictionary including config info
    :param tbrn: path to be removed, like "tb:KITTI:seg1".
    :param is_recursive: whether remove directories recursively
    :param force_delete: sensor and its objects will also be deleted if True,
      else only segment with no sensor can be deleted.
    """
    gas = _gas(**obj)
    info = TBRN(tbrn=tbrn)
    dataset = gas.get_dataset(info.dataset_name)

    if info.type == TBRNType.DATASET:
        if not is_recursive:
            click.echo("Error: please use -r option to remove the whole dataset", err=True)
            sys.exit(1)
        segment_names = dataset.list_segments()
        dataset.delete_segments(segment_names, force_delete)
        return

    if info.type == TBRNType.SEGMENT:
        if not is_recursive:
            click.echo("Error: please use -r option to remove the whole segment", err=True)
            sys.exit(1)
        dataset.delete_segments(info.segment_name, force_delete)
        return

    if info.type == TBRNType.NORMAL_FILE:
        if not is_recursive and info.remote_path.endswith("/"):
            click.echo("Error: please use -r option to remove recursively", err=True)
            sys.exit(1)

        segment = dataset.get_segment(info.segment_name)
        filter_data = list(_filter_data(segment.list_data(), info.remote_path, is_recursive))
        if not filter_data:
            echo_info = "file or directory" if is_recursive else "file"
            click.echo(f'Error: no such {echo_info} "{tbrn}" ', err=True)
            sys.exit(1)
        segment.delete_data(filter_data)
        return

    click.echo(f'"{tbrn}" is an invalid path to remove', err=True)
    sys.exit(1)


@cli.command()
@click.argument("access_key", type=str, default="")
@click.argument("url", type=str, default="")
@click.pass_obj
def config(obj: Dict[str, str], access_key: str, url: str) -> None:
    """Configure the accessKey (and url) of gas.\f

    :param obj: a dictionary including config info
    :param access_key: the accessKey of gas to write into config file
    :param url: the url of gas to write into config file
    """
    config_file = _config_filepath()
    config_parser = ConfigParser()
    config_parser.read(config_file)

    if not access_key:
        for config_name in config_parser.sections():
            click.echo(f"[{config_name}]")
            for key, value in config_parser[config_name].items():
                click.echo(f"{key} = {value}")
        return

    if not access_key.startswith(_ACCESSKEY_PREFIX):
        click.echo(f'Error: AccessKey should start with "{_ACCESSKEY_PREFIX}"', err=True)
        sys.exit(1)

    config_name = obj["config_name"]
    if config_name not in config_parser:
        config_parser.add_section(config_name)

    config_parser[config_name]["accessKey"] = access_key
    if url:
        config_parser[config_name]["url"] = url
    else:
        config_parser.remove_option(config_name, "url")

    with open(config_file, "w") as file:
        config_parser.write(file)

    click.echo(f"Success!\nConfiguration has been written into: {config_file}")


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""client module."""

from .exceptions import (
    GASDatasetError,
    GASDatasetTypeError,
    GASException,
    GASFrameError,
    GASOSSError,
    GASPathError,
    GASRequestError,
    GASResponseError,
    GASResponseException,
    GASTensorBayError,
)
from .gas import GAS

__all__ = [
    "GAS",
    "GASDatasetError",
    "GASDatasetTypeError",
    "GASException",
    "GASFrameError",
    "GASOSSError",
    "GASPathError",
    "GASRequestError",
    "GASResponseError",
    "GASResponseException",
    "GASTensorBayError",
]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Box2D, Box3D"""

from typing import Dict, List, Optional, Sequence, TypeVar, Union, overload

from .quaternion import Quaternion
from .transform import Transform3D
from .vector import Vector2D, Vector3D

T = TypeVar("T", bound="Box3D")  # pylint: disable=invalid-name


class Box2D(Sequence[float]):
    """Contain the definition of 2D bounding box and some related operations.

    :param args: Union[None, float, Sequence[float]],
        box = Box2D()
        box = Box2D(10, 20, 30, 40)
        box = Box2D([10, 20, 30, 40])
    :param loads: [
        {
            "x": ...
            "y": ...
        },
        {
            "x": ...
            "y": ...
        }
    ]
    :param x: X coordinate of the top left vertex of the box
    :param y: Y coordinate of the top left vertex of the box
    :param width: Length along the x axis
    :param height: Length along the y axis
    :raise TypeError: When input params do not meet the requirement
    """

    def __init__(
        self,
        *args: Union[None, float, Sequence[float]],
        loads: Optional[List[Dict[str, float]]] = None,
        x: Optional[float] = None,
        y: Optional[float] = None,
        width: Optional[float] = None,
        height: Optional[float] = None,
    ) -> None:
        # pylint: disable=invalid-name

        if loads:
            tl_x, tl_y = loads[0]["x"], loads[0]["y"]
            br_x, br_y = loads[1]["x"], loads[1]["y"]

        elif x is not None or y is not None or width is not None or height is not None:
            try:
                tl_x, tl_y = x, y  # type: ignore[assignment]
                br_x, br_y = x + width, y + height  # type: ignore[operator]
            except TypeError as error:
                raise TypeError(
                    "Require x, y, width, height keyword arguments to construct a 2D box."
                ) from error

        else:
            arg = args[0] if len(args) == 1 else args
            if not arg:
                tl_x, tl_y = 0, 0
                br_x, br_y = 0, 0
            elif isinstance(arg, Sequence) and len(arg) == 4:  # pylint: disable=W1116
                tl_x, tl_y = arg[0:2]  # type: ignore[assignment]
                br_x, br_y = arg[2:4]  # type: ignore[assignment]
            else:
                raise ValueError("Require 4 coordinates to construct a 2D box.")

        if tl_x > br_x or tl_y > br_y:
            tl = Vector2D()
            br = Vector2D()
        else:
            tl = Vector2D(tl_x, tl_y)
            br = Vector2D(br_x, br_y)
        self._data = (tl, br)

    def dumps(self) -> List[Dict[str, float]]:
        """Dump a 2D box as a list.

        :return: a list containing vertex coordinates of the box
        """
        return [{"x": self.tl.x, "y": self.tl.y}, {"x": self.br.x, "y": self.br.y}]

    def __repr__(self) -> str:
        """Print as Box2D(x1, y1, x2, y2)

        :return: a string like "Box2D(x1, y1, x2, y2)"
        """
        return f"{self.__class__.__name__}({self.tl.x}, {self.tl.y}, {self.br.x}, {self.br.y})"

    # pylint: disable=invalid-name
    def __and__(self, other: "Box2D") -> "Box2D":
        """Calculate the intersect box of two boxes.

        :param other: the other box
        :return: the intersect box of the two boxes
        """
        x1 = max(self.tl.x, other.tl.x)
        x2 = min(self.br.x, other.br.x)
        y1 = max(self.tl.y, other.tl.y)
        y2 = min(self.br.y, other.br.y)
        return Box2D(x1, y1, x2, y2)

    @overload
    def __getitem__(self, index: int) -> float:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[float]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[Sequence[float], float]:
        return (tuple(self.tl) + tuple(self.br)).__getitem__(index)

    def __len__(self) -> int:
        return 4

    @property
    def tl(self) -> Vector2D:  # pylint: disable=invalid-name
        """Get the top left point.

        :return: the top left point
        """
        return self._data[0]

    @property
    def br(self) -> Vector2D:  # pylint: disable=invalid-name
        """Get the bottom right point.

        :return: the bottom right point
        """
        return self._data[1]

    @property
    def width(self) -> float:
        """Get the width of the 2d box"""
        return self.br.x - self.tl.x

    @property
    def height(self) -> float:
        """Get the height of the 2d box"""
        return self.br.y - self.tl.y

    def area(self) -> float:
        """Get the area of the 2d box"""
        return self.width * self.height

    @staticmethod
    def iou(box1: "Box2D", box2: "Box2D") -> float:
        """Calculate the intersection over union of two 2d boxes.

        :param box1: a 2d box
        :param box2: a 2d box
        :return: intersection over union between the two input boxes
        """
        area1 = box1.area()
        area2 = box2.area()
        intersect_box = box1 & box2
        intersect = intersect_box.area()
        union = area1 + area2 - intersect
        return intersect / union


class Box3D:
    """Contain the definition of 3D bounding box and some related operations.

    :param transform: A Transform3D object or a 4x4 or 3x4 transfrom matrix
    :param translation: Translation in a sequence of [x, y, z]
    :param rotation: Rotation in a sequence of [w, x, y, z] or 3x3 rotation matrix or `Quaternion`
    :param size: Size in a sequence of [x, y, z]
    :param loads: {
        "translation": {
            "x": ...
            "y": ...
            "z": ...
        },
        "rotation": {
            "w": ...
            "x": ...
            "y": ...
            "z": ...
        },
        "size": {
            "x": ...
            "y": ...
            "z": ...
        }
    }
    :param kwargs: Other parameters to initialize rotation of the transform
    """

    def __init__(
        self,
        transform: Transform3D.TransformType = None,
        *,
        translation: Optional[Sequence[float]] = None,
        rotation: Quaternion.ArgsType = None,
        size: Optional[Sequence[float]] = None,
        loads: Optional[Dict[str, Dict[str, float]]] = None,
        **kwargs: Quaternion.KwargsType,
    ) -> None:
        if loads:
            self._size = Vector3D(loads=loads["size"])
            self._transform = Transform3D(loads=loads)
            return

        self._transform = Transform3D(
            transform, translation=translation, rotation=rotation, loads=None, **kwargs
        )
        self._size = Vector3D(size)

    def dumps(self) -> Dict[str, Dict[str, float]]:
        """Dump the 3d box as a dictionary.

        :return: a dictionary containing translation, rotation and size info
        """
        box3d_dict = self._transform.dumps()
        box3d_dict["size"] = self.size.dumps()
        return box3d_dict

    def __repr__(self) -> str:
        translation = self.translation
        rotation = self.rotation
        size = self.size

        return (
            f"{self.__class__.__name__}("
            f"\n  Translation({translation.x}, {translation.y}, {translation.z}),"
            f"\n  Rotation({rotation.w}, {rotation.x}, {rotation.y}, {rotation.z}),"
            f"\n  Size({size.x}, {size.y}, {size.z}),"
            "\n)"
        )

    def __rmul__(self: T, other: Transform3D) -> T:
        if isinstance(other, Transform3D):
            box: T = object.__new__(self.__class__)
            box._transform = other * self._transform
            box._size = self._size
            return box

        return NotImplemented  # type: ignore[unreachable]

    @property
    def translation(self) -> Vector3D:
        """Get the translation of the 3d box by property."""
        return self._transform.translation

    @property
    def rotation(self) -> Quaternion:
        """Get the rotation of the 3d box by property."""
        return self._transform.rotation

    @property
    def transform(self) -> Transform3D:
        """Get the transform of the 3d box by property."""
        return self._transform

    @property
    def size(self) -> Vector3D:
        """Get the size of the 3d box by property."""
        return self._size

    def volume(self) -> float:
        """Get the volume of the 3d box."""
        return self.size.x * self.size.y * self.size.z

    @classmethod
    def iou(cls, box1: "Box3D", box2: "Box3D", angle_threshold: float = 5) -> float:
        """Calculate the iou between two 3d boxes.

        :param box1: a 3d box
        :param box2: a 3d box
        :param angle_threshold: the threshold of the relative angles between two input 3d boxes,
        in degree
        :return: the iou of the two 3d boxes
        """
        box2 = box1.transform.inverse() * box2
        if abs(box2.rotation.degrees) > angle_threshold:
            return 0

        intersect_size = [
            cls._line_intersect(*args) for args in zip(box1.size, box2.size, box2.translation)
        ]
        intersect = intersect_size[0] * intersect_size[1] * intersect_size[2]
        union = box1.volume() + box2.volume() - intersect
        return intersect / union

    @staticmethod
    def _line_intersect(length1: float, length2: float, midpoint_distance: float) -> float:
        """Calculate the intersect length between two parallel lines.

        :param length1: the length of line1
        :param length2: the length of line2
        :param midpoint_distance: the distance between midpoints of the two lines
        :return: the intersect length between line1 and line2
        """
        line1_min = -length1 / 2
        line1_max = length1 / 2
        line2_min = -length2 / 2 + midpoint_distance
        line2_max = length2 / 2 + midpoint_distance
        intersect_length = min(line1_max, line2_max) - max(line1_min, line2_min)
        return intersect_length if intersect_length > 0 else 0



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file contains checkhealth related functions."""

from typing import Union

from .dataset import Dataset, FusionDataset


def checkhealth(dataset: Union[Dataset, FusionDataset]) -> bool:
    pass



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Transform3D.
"""

from typing import Dict, Optional, Sequence, Type, TypeVar, Union, overload

import numpy as np

from .quaternion import Quaternion
from .vector import Vector3D

T = TypeVar("T", bound="Transform3D")  # pylint: disable=invalid-name


class Transform3D:
    """A class used to represent a transformation in 3D.

    :param transform: A Transform3D object or a 4x4 or 3x4 transfrom matrix
    :param translation: Translation in a sequence of [x, y, z]
    :param rotation: Rotation in a sequence of [w, x, y, z] or 3x3 rotation matrix or Quaternion
    :param loads: A dictionary containing the translation and the rotation
    :param kwargs: Other parameters to initialize rotation of the transform
    :raises ValueError: When the shape of the input matrix is wrong
    """

    MatrixType = Union[None, Sequence[Sequence[float]], np.ndarray]
    TransformType = Union[None, "Transform3D", Sequence[Sequence[float]], np.ndarray]

    def __init__(
        self,
        transform: TransformType = None,
        *,
        translation: Optional[Sequence[float]] = None,
        rotation: Quaternion.ArgsType = None,
        loads: Optional[Dict[str, Dict[str, float]]] = None,
        **kwargs: Quaternion.KwargsType,
    ) -> None:
        if loads:
            self._translation = Vector3D(loads=loads["translation"])
            self._rotation = Quaternion(loads=loads["rotation"])
            return

        if transform is not None:
            if isinstance(transform, Transform3D):
                self._translation = transform.translation
                self._rotation = transform.rotation
                return

            if isinstance(transform, Sequence):  # pylint: disable=W1116
                transform = np.array(transform)
            if transform.shape != (3, 4) and transform.shape != (4, 4):
                raise ValueError("The shape of input transform matrix must be 3x4 or 4x4.")

            self._translation = Vector3D(transform[0, 3], transform[1, 3], transform[2, 3])
            self._rotation = Quaternion(transform)
            return

        self._translation = Vector3D(translation)
        self._rotation = Quaternion(rotation, loads=None, **kwargs)

    def dumps(self) -> Dict[str, Dict[str, float]]:
        """Dumps the 3D transform as a dictionary.

        :return: A dictionary containing rotation and translation information of the transform3D
        """
        return {
            "translation": self._translation.dumps(),
            "rotation": self._rotation.dumps(),
        }

    @overload
    def __mul__(self: T, other: T) -> T:
        ...

    @overload
    def __mul__(self: T, other: Quaternion) -> T:
        ...

    @overload
    def __mul__(self: T, other: Sequence[float]) -> Vector3D:
        ...

    def __mul__(self: T, other: Union[T, Quaternion, Sequence[float]]) -> Union[T, Vector3D]:

        if isinstance(other, Sequence):  # pylint: disable=W1116
            return self._translation + self._rotation.rotate(other)

        if isinstance(other, Quaternion):
            return self._create(self._translation, self._rotation * other)

        if isinstance(other, Transform3D):
            return self._create(self * other.translation, self._rotation * other.rotation)

        return NotImplemented  # type: ignore[unreachable]

    def __rmul__(self: T, other: Quaternion) -> T:
        if isinstance(other, Quaternion):
            return self._create(other * self._translation, other * self._rotation)

        return NotImplemented  # type: ignore[unreachable]

    def __repr__(self) -> str:
        translation = self._translation
        rotation = self._rotation
        return (
            f"{self.__class__.__name__}("
            f"\n  Translation({translation.x}, {translation.y}, {translation.z}),"
            f"\n  Rotation({rotation.w}, {rotation.x}, {rotation.y}, {rotation.z}),"
            "\n)"
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._translation.__eq__(other.translation) and self._rotation == other.rotation

        return False

    @property
    def translation(self) -> Vector3D:
        """Returns translation of the 3D transform.

        :return: Translation in Vector3D
        """
        return self._translation

    def set_translation(self, *args: Union[float, Sequence[float]], **kwargs: float) -> None:
        """Set the translation of the transform.

        :param args: Coordinates of the translation vector
        :param kwargs: keyword-only argument to set different dimension for translation vector
            transform.set_translation(x=1, y=2, z=3)
        """
        self._translation = Vector3D(*args, loads=None, **kwargs)

    @property
    def rotation(self) -> Quaternion:
        """Returns rotation of the 3D transform.

        :return: rotation in Quaternion
        """
        return self._rotation

    def set_rotation(
        self,
        *args: Union[Quaternion.ArgsType, float],
        **kwargs: Quaternion.KwargsType,
    ) -> None:
        """Set the rotation of the transform.

        :param args: Coordinates of the Quaternion
        :param kwargs: keyword-only argument to the Quaternion
        """
        self._rotation = Quaternion(*args, loads=None, **kwargs)

    def as_matrix(self) -> np.ndarray:
        """Return the transform as a 4x4 transform matrix.

        :return: A 4x4 numpy array representing the transform matrix
        """
        matrix = np.eye(4)
        matrix[:3, 3] = self._translation
        matrix[:3, :3] = self._rotation.as_matrix()
        return matrix

    def inverse(self: T) -> T:
        """Return the inverse of the transform.

        :return: A Transform3D object representing the inverse of this Transform3D
        """
        rotation = self._rotation.inverse()
        translation = rotation.rotate(-self._translation)
        return self._create(translation, rotation)

    @classmethod
    def _create(cls: Type[T], translation: Vector3D, rotation: Quaternion) -> T:
        transform: T = object.__new__(cls)
        transform._translation = translation  # pylint: disable=protected-access
        transform._rotation = rotation  # pylint: disable=protected-access
        return transform



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

""" sensor related classes """

from .intrinsics import CameraIntrinsics, CameraMatrix, DistortionCoefficients
from .sensor import Camera, FisheyeCamera, Lidar, Radar, Sensor, SensorType

__all__ = [
    "Camera",
    "DistortionCoefficients",
    "CameraIntrinsics",
    "CameraMatrix",
    "FisheyeCamera",
    "Lidar",
    "Radar",
    "Sensor",
    "SensorType",
]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Basic, Vector, Vector2D and Vector3D.
"""

from typing import Dict, Optional, Sequence, Tuple, Type, TypeVar, Union, overload

T = TypeVar("T", bound="Vector")  # pylint: disable=invalid-name


class Vector(Sequence[float]):
    """A class used as base class of Vector2D and Vector3D, extending the :class `Sequence`.

    :param args: Coordinates of the vector
    :param loads: A dicitionary containing coordinates of a vector
    :param kwargs: float
        vector2d = Vector(x=1, y=2)
        vector3d = Vector(x=1, y=2, z=3)
    :raises ValueError: When the dimension of the input parameters is not correct
    :return: The created Vector2D or Vector3D class
    """

    _DIMENSION: Optional[int] = None

    _data: Tuple[float, ...]

    def __new__(
        cls: Type[T],
        *args: Union[None, float, Sequence[float]],
        loads: Optional[Dict[str, float]] = None,
        **kwargs: float,
    ) -> T:
        data = cls._process_args(*args, loads=loads, **kwargs)
        ReturnType: Type["Vector"] = cls
        if not cls._DIMENSION:
            if len(data) == Vector2D._DIMENSION:
                ReturnType = Vector2D
            elif len(data) == Vector3D._DIMENSION:
                ReturnType = Vector3D
            else:
                raise ValueError("Require 2 or 3 dimensional data to construct a Vector.")
        elif not data:
            data = (0,) * cls._DIMENSION
        elif len(data) != cls._DIMENSION:
            raise ValueError(f"Wrong dimension of input data. Require {cls._DIMENSION} dimension.")

        obj: T = object.__new__(ReturnType)
        obj._data = data
        return obj

    def __repr__(self) -> str:
        """Returns the representation of the vector.

        :returns: Vector representation in str
        """
        return f"{self.__class__.__name__}{self._data}"

    def __add__(self, other: Sequence[float]) -> T:
        """Calculate the sum of vector and other sequence.

        :raises ValueError: When the vectors to be added have different dimensions
        :returns: The sum vector of adding two vectors
        """
        if not isinstance(other, Sequence):  # pylint: disable=W1116
            return NotImplemented

        if len(self._data) != len(other):
            raise ValueError("Vectors to be added must have the same dimension")
        result: T = object.__new__(self.__class__)
        result._data = tuple(i + j for i, j in zip(self._data, other))
        return result

    def __radd__(self, other: Sequence[float]) -> T:
        """Calculate the sum of other sequence and the vector.

        :returns: The sum vector of adding two vectors
        """
        return self.__add__(other)

    def __neg__(self) -> T:
        result: T = object.__new__(self.__class__)
        result._data = tuple(-coordinate for coordinate in self._data)
        return result

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._data.__eq__(other._data)

        return False

    @overload
    def __getitem__(self, index: int) -> float:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[float]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[Sequence[float], float]:
        """Returns the corresponding items.

        :param index: An int index or a slice for indexing
        :return: A float or a Sequence of float corresponding to the input index
        """
        return self._data.__getitem__(index)

    def __len__(self) -> int:
        """Returns the dimension of the vector.

        :returns: Number of dimension of the vector.
        """
        return self._data.__len__()

    @staticmethod
    def _process_args(
        *args: Union[None, float, Sequence[float]],
        loads: Optional[Dict[str, float]] = None,
        **kwargs: float,
    ) -> Tuple[float, ...]:
        if loads:
            kwargs = loads

        if kwargs:
            if "z" in kwargs:
                return (kwargs["x"], kwargs["y"], kwargs["z"])
            return (kwargs["x"], kwargs["y"])

        data = args[0] if len(args) == 1 else args
        if not data:
            return ()

        if isinstance(data, Sequence):  # pylint: disable=W1116
            return tuple(data)  # type: ignore[arg-type]

        raise ValueError("Require 2 or 3 dimensional data to construct a vector.")


class Vector2D(Vector):
    """A class used to represent 2D vector extending the :class `Vector`."""

    _DIMENSION = 2

    @property
    def x(self) -> float:  # pylint: disable=invalid-name
        """Returns the x coordinate of the vector.

        :return: x coordinate in float type
        """
        return self._data[0]

    @property
    def y(self) -> float:  # pylint: disable=invalid-name

        """Returns the y coordinate of the vector.

        :return: y coordinate in float type
        """
        return self._data[1]

    def dumps(self) -> Dict[str, float]:
        """Dumps the vector as a dictionary.

        :return: A dictionary containing the vector coordinate.
        """
        return {"x": self._data[0], "y": self._data[1]}


class Vector3D(Vector):
    """A class used to represent 3D vector extending the :class `Vector`."""

    _DIMENSION = 3

    @property
    def x(self) -> float:  # pylint: disable=invalid-name

        """Returns the x coordinate of the vector.

        :return: x coordinate in float type
        """
        return self._data[0]

    @property
    def y(self) -> float:  # pylint: disable=invalid-name

        """Returns the y coordinate of the vector.

        :return: y coordinate in float type
        """
        return self._data[1]

    @property
    def z(self) -> float:  # pylint: disable=invalid-name

        """Returns the z coordinate of the vector.

        :return: z coordinate in float type
        """
        return self._data[2]

    def dumps(self) -> Dict[str, float]:
        """Dumps the vector as a dictionary.

        :return: A dictionary containing the vector coordinate.
        """
        return {"x": self._data[0], "y": self._data[1], "z": self._data[2]}



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines the class Labelset.
"""

from typing import Any, Dict, Generator, Iterable, List, Optional, Set, Tuple

from ..dataset import Data, Frame, FusionSegment, Segment
from ..label import CategoryInfo, LabelTables, LabelType
from ..utility import TBRN, NameSortedDict, TBRNType
from .path import SegmentNameWrapper
from .requests import Client, multithread_upload

_FULL_DATA_LIST_TYPE = 2


class LabelSegmentClientBase:  # pylint: disable=too-few-public-methods
    """This class defines the concept of a labelset segment client and some opertions on it.

    :param name: Segment name, unique for a labelset
    :param labelset_id: Labelset ID
    :param dataset_name: Dataset name
    :param client: The client used for sending request to TensorBay
    """

    def __init__(self, name: str, labelset_id: str, dataset_name: str, client: Client) -> None:
        self._name = name
        self._labelset_id = labelset_id
        self._dataset_name = dataset_name
        self._client = client
        self._name_wrapper = SegmentNameWrapper(name)

    def get_segment_name(self) -> str:
        """Return the name of this labelset segment."""
        return self._name

    def _get_url(self, tbrn: TBRN) -> str:
        """Get url of specific remote_path from labelset

        :param remote_path: The remote_path of the data
        :param sensor_name: The name of the of the data sensor
        :return: The url of the input remote_path and sensor_name
        """
        remote_path = tbrn.remote_path
        sensor_name = tbrn.sensor_name if tbrn.type == TBRNType.FUSION_FILE else ""
        post_data = {
            "labelSetId": self._labelset_id,
            "listType": _FULL_DATA_LIST_TYPE,
            "needLabel": False,
            "objectPath": self._name_wrapper.get_object_path(remote_path, sensor_name),
            "outputVersion": "v1.0.2",
        }
        response = self._client.labelset_post("getUrlAndLabel", post_data)
        return response["url"]  # type: ignore[no-any-return]

    def _upload_label(
        self,
        data: Data,
        sensor_name: str = "",
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Upload label to the labelset.

        :param data: The Data contains the labels to be uploaded
        :param sensor_name: Sensor name of the Data
        :param metadata: Some additional data of the label
        """
        remote_path = data.remote_path

        post_data: Dict[str, Any] = {}
        post_data["ObjectPath"] = self._name_wrapper.get_object_path(remote_path, sensor_name)
        post_data["labelValues"] = data.dump_labels()
        post_data["labelSetId"] = self._labelset_id
        if metadata:
            post_data["labelMeta"] = metadata
        self._client.labelset_post("putLabel", post_data)


class LabelSegmentClient(LabelSegmentClientBase):
    """A class used to represent a segment client of a labelset."""

    def upload_data_object(self, data: Data) -> None:
        """Upload the input data's label to the labelset segment.

        :param data: The `Data` object which contains the labels to be uploaded
        """
        self._upload_label(data)

    def list_data(self) -> List[str]:
        """List all data remote path in a labelset segment.

        :return: A list of data path
        """
        post_data = {
            "labelSetId": self._labelset_id,
            "listType": _FULL_DATA_LIST_TYPE,
            "segmentName": self._name_wrapper.post_segment_name,
        }
        objects = self._client.labelset_post("listObjectsBySegment", post_data)["objects"]
        return [self._name_wrapper.get_remote_path(info["objectName"]) for info in objects]

    def list_data_objects(self) -> List[Data]:
        """List all `Data` object in a labelset segment.

        :return: A list of `Data` object
        """
        labels = self._list_data_summaries()

        data_list = []
        for label in labels:
            remote_path = self._name_wrapper.get_remote_path(label["objectPath"])
            tbrn = TBRN(self._dataset_name, self._name, remote_path=remote_path)
            data = Data(tbrn, url_getter=self._get_url)
            data.load_labels(label["values"])
            data_list.append(data)

        return data_list

    def _list_data_summaries(self) -> List[Dict[str, Any]]:
        post_data = {
            "labelSetId": self._labelset_id,
            "listType": _FULL_DATA_LIST_TYPE,
            "segmentName": self._name_wrapper.post_segment_name,
            "outputVersion": "v1.0.2",
            "pageSize": -1,
            "projection": {
                "objectpath": 1,
                "values": 1,
            },
        }
        response = self._client.labelset_post("listLabelsBySegment", post_data)
        return response["labels"]  # type: ignore[no-any-return]

    def _list_uploaded_data(self) -> Set[str]:
        uploaded_set = set()
        labels = self._list_data_summaries()
        for label in labels:
            if not label["values"]:
                continue
            remote_path = self._name_wrapper.get_remote_path(label["objectPath"])
            uploaded_set.add(remote_path)

        return uploaded_set


class FusionLabelSegmentClient(LabelSegmentClientBase):
    """A class used to represent a segment client of a fusion labelset."""

    def upload_frame_object(self, frame: Frame) -> None:
        """Upload the input frame's label to the fusion labelset segment.

        :param frame: The `Frame` object which contains the labels to be uploaded
        """
        for sensor_name, data in frame.items():
            self._upload_label(data, sensor_name)

    def list_frame_objects(self) -> List[Frame]:
        """List all frames in the labelset segment without the label info.

        :return: A list `Frame` object
        """
        summaries = self._list_frame_summaries()
        frames = []
        for frame_info in summaries:
            frame = Frame()
            for data_info in frame_info:
                sensor_name = data_info["sensorName"]
                object_path = data_info["objectPath"]
                remote_path = self._name_wrapper.get_remote_path(object_path, sensor_name)
                data = Data(remote_path, timestamp=data_info.get("timestamp", None))
                data.load_labels(data_info["label"]["values"])
                frame[sensor_name] = data

            frames.append(frame)

        return frames

    def _list_frame_summaries(self) -> List[List[Dict[str, Any]]]:
        post_data = {
            "labelSetId": self._labelset_id,
            "listType": _FULL_DATA_LIST_TYPE,
            "pageSize": -1,
            "needLabel": True,
            "outputVersion": "v1.0.2",
            "projection": {
                "objectpath": 1,
                "values": 1,
            },
            "segmentName": self._name_wrapper.post_segment_name,
        }
        response = self._client.labelset_post("listFrameObjects", post_data)
        return response["frameObjects"]  # type: ignore[no-any-return]

    def _list_uploaded_data(self) -> Set[Tuple[str, str]]:
        uploaded_set = set()
        summaries = self._list_frame_summaries()
        for frame_info in summaries:
            for data_info in frame_info:
                if not data_info["label"]["values"]:
                    continue
                sensor_name = data_info["sensorName"]
                object_path = data_info["objectPath"]
                remote_path = self._name_wrapper.get_remote_path(object_path, sensor_name)
                uploaded_set.add((sensor_name, remote_path))

        return uploaded_set


class LabelsetClientBase:
    """A base class which respresents a client of labelset.

    :param labelset_id: The ID of the labelset
    :param dataset_name: The name of the dataset
    :param client: The client used for sending request to TensorBay
    """

    TYPE_GROUND_TRUTH = 3
    _PUBLISH_STATUS = 2
    _STATISTICSED_STATUS = 4
    _TASK_TYPE = {
        LabelType.CLASSIFICATION: {"code": 7, "nameEn": "2D CLASSIFICATION"},
        LabelType.BOX2D: {"code": 10, "nameEn": "2D BOX"},
        LabelType.BOX3D: {"code": 4, "nameEn": "3D BOX"},
        LabelType.POLYGON: {"code": 22, "nameEn": "2D POLYGON"},
        LabelType.POLYLINE: {"code": 8, "nameEn": "2D POLYLINE"},
    }
    _DELIMITER = "."

    def __init__(self, labelset_id: str, dataset_name: str, client: Client) -> None:
        self._labelset_id = labelset_id
        self._dataset_name = dataset_name
        self._client = client

    def get_labelset_id(self) -> str:
        """Return the id of the labelset.

        :return: The id of the labelset
        """
        return self._labelset_id

    def upload_label_tables(self, label_tables: LabelTables) -> None:
        """Upload label tables to the labelset.

        :param label_tables: The label tables to be uploaded
        """
        metadata = LabelsetClientBase.create_metadata(label_tables)
        if not metadata:
            return

        post_data = {
            "labelSetId": self._labelset_id,
            "meta": metadata,
        }
        self._client.labelset_post("updateLableSet", post_data)

    def list_segments(self) -> List[str]:
        """List all segment names in a labelset.

        :return: A list of segment names
        """
        post_data = {
            "labelSetId": self._labelset_id,
            "listType": _FULL_DATA_LIST_TYPE,
            "pageSize": -1,
        }
        segment_names = self._client.labelset_post("listSegments", post_data)["segments"]
        return [SegmentNameWrapper.get_local_segment_name(name) for name in segment_names]

    def publish(self) -> None:
        """Publish the labelset."""
        post_data = {"labelSetId": self._labelset_id, "status": LabelsetClientBase._PUBLISH_STATUS}
        self._client.labelset_post("updateLabelSetStatus", post_data)

    def is_published(self) -> bool:
        """Gheck whether the labelset is published.

        :return: Return `True` if the lableset is publish, otherwise return `False`
        """
        post_data = {"id": self._labelset_id, "projection": {"status": 1}}
        data = self._client.labelset_post("listLabelSets", post_data)
        status = data["labelSets"][0]["labelSet"]["status"]
        return status == LabelsetClientBase._STATISTICSED_STATUS  # type: ignore[no-any-return]

    @staticmethod
    def create_metadata(label_tables: LabelTables) -> Dict[str, Any]:
        """Return metadata created from the label tables.

        :return: A dict in metadata format
        """
        task_types = []
        for label_type, label_table in label_tables.items():
            task_type = LabelsetClientBase._TASK_TYPE[label_type].copy()
            task_type["is_tracking"] = label_table.is_tracking
            if label_table.categories:
                task_type["categories"] = LabelsetClientBase._create_categories_meta(
                    label_table.categories
                )
            task_types.append(task_type)

        return {"taskTypes": task_types} if task_types else {}

    @staticmethod
    def _create_categories_meta(categories: NameSortedDict[CategoryInfo]) -> List[Dict[str, Any]]:
        """Create the category metadata with parent-child relationship.

        :param categories: The input categories in a NameSortedDict
        :return: The created category metadata list
        """
        category_tree = []
        tree_index: Dict[str, Any] = {}
        for category in categories:
            category_info = {"nameEn": category}
            while True:
                i = category.rfind(LabelsetClientBase._DELIMITER)
                if i == -1:
                    category_tree.append(category_info)
                    break

                category = category[:i]
                if category in tree_index:
                    tree_index[category].append(category_info)
                    break

                tree_index[category] = [category_info]
                category_info = {"nameEn": category, "subcategories": tree_index[category]}
        return category_tree


class LabelsetClient(LabelsetClientBase):
    """A class which respresents a client of a labelset."""

    def get_segment(self, name: str) -> LabelSegmentClient:
        """Get a labelset segment according to the input name.

        :param name: The name of the requested labelset segment
        :return: The requested labelset segment client
        """

        return LabelSegmentClient(name, self._labelset_id, self._dataset_name, self._client)

    def upload_segment_object(
        self,
        segment: Segment,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> LabelSegmentClient:
        """Upload a `Segment` to the labelset,
        This function will upload all labels contains in the input `Segment` object,

        :param segment: The `Segment` object contains the labels needs to be uploaded
        :param jobs: The number of the max workers in multithread upload
        :return: The client used for uploading the label in the `Segment`
        """
        segment_client = self.get_segment(segment.name)

        segment_filter: Iterable[Data]
        if skip_uploaded_files:
            done_set = segment_client._list_uploaded_data()  # pylint: disable=protected-access
            segment_filter = filter(lambda data: data.remote_path not in done_set, segment)
        else:
            segment_filter = segment

        multithread_upload(segment_client.upload_data_object, segment_filter, jobs=jobs)
        return segment_client


class FusionLabelsetClient(LabelsetClientBase):
    """A class which respresents a client of a fusion labelset."""

    def get_segment(self, name: str) -> FusionLabelSegmentClient:
        """Get a client of fusion labelset segment according to the input name.

        :param name: The name of the requested fusion labelset segment
        :return: The requested client of fusion labelset segment
        """
        return FusionLabelSegmentClient(name, self._labelset_id, self._dataset_name, self._client)

    def upload_segment_object(
        self,
        segment: FusionSegment,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> FusionLabelSegmentClient:
        """Upload a `FusionSegment` to the labelset,
        This function will upload all labels contains in the input `FusionSegment` object,

        :param segment: The `FusionSegment` object contains the labels needs to be uploaded
        :param jobs: The number of the max workers in multithread upload
        :return: The client used for uploading the label in the `FusionSegment`
        """
        segment_client = self.get_segment(segment.name)

        if skip_uploaded_files:
            done_set = segment_client._list_uploaded_data()  # pylint: disable=protected-access
        else:
            done_set = set()

        segment_filter = self._frame_data_generator(segment, done_set)

        multithread_upload(
            lambda args: segment_client._upload_label(  # pylint: disable=protected-access
                args[1], args[0]
            ),
            segment_filter,
            jobs=jobs,
        )

        return segment_client

    @staticmethod
    def _frame_data_generator(
        segment: FusionSegment,
        uploaded_set: Set[Tuple[str, str]],
    ) -> Generator[Tuple[str, Data], None, None]:
        for frame in segment:
            for sensor_name, data in frame.items():
                if (sensor_name, data.remote_path) in uploaded_set:
                    continue

                yield (sensor_name, data)
